/*
** Module   : EnvironmentOps.java
** Abstract : Progress 4GL compatible environment related methods
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050628   @21564 Created initial version, supporting common
**                           global variables that are called built-in
**                           functions in Progress but which take no
**                           parameters, no empty parens. These also
**                           are read-only (they cannot be assigned).
** 002 GES 20050729   @21921 Added getSourceName() and modified the
**                           default propath.
** 003 ECF 20051214   @23753 Removed getCurrentDatabaseName(). This was a
**                           placeholder for P2J's DBNAME replacement
**                           implementation. This is now provided in the
**                           ConnectionManager class instead.
** 004 GES 20060123   @24034 Move to a directory based implementation for
**                           the Progress compatibility global vars.
** 005 GES 20060125   @24071 Shift to SourceNameMapper backed stack
**                           processing.
** 006 EVL 20060713   @28035 Changing the approach to get the source file
**                           name from index of the Progress procedure on
**                           the stack.
** 007 GES 20070321   @32506 Added support for assigning the PROPATH.
** 008 GES 20071214   @36369 Fix for NPE in getSourceName().
** 009 GES 20080123   @36902 Better propath and separator defaults.
** 010 GES 20080321   @37628 Added SESSION:BATCH-MODE support.
** 011 GES 20080414   @38014 Added SESSION:CHARSET support.
** 012 GES 20080924   @39940 Mask off inner classes from program-name
**                           processing.
** 013 GES 20090512   @42147 Minor optimization for program-name.
** 014 ECF 20101005          Added getTempDirectory method.
** 015 GES 20110713          Changed default 4GL version string to "10.2B".
** 016 CS  20121104          Implement PROPATH assignment.
** 017 EVL 20130108          Fixing the PROPATH handling. We have to substitute
**                           system path separator with comma when getting
**                           PROPATH value. Also the default path and file
**                           separators value is chaged to return the current
**                           OS dependent one from File.pathSeparator and
**                           File.separator. Also the getLegacyCaseSensitive()
**                           method has been changed to set the default value
**                           depending on the current OS.
** 018 EVL 20130128          Adding support for GET-KEY-VALUE, PUT-KEY-VALUE,
**                           USE, LOAD, UNLOAD statements. The COLOR and FONT
**                           options for PUT-KEY-VALUE is not yet supported.
**                           The corresponting Java methods are just stubs to
**                           be implemented later.
** 019 VMN 20130219          Added support for attributes
**                           SESSION:FIRST-BUFFER (KW_FIRST_BU)
** 020 CA  20130222          Removed getFirstBuffer (stub is in SessionUtils).
** 021 EVL 20130529          Changing the setSearchPath method to take into account the current
**                           PROPATH value while setting up the new one. The new item is adding
**                           in the head of the string replacing the variable part of the PROPATH
**                           value. Also the methods splitPath and fixupPropath were modified to
**                           consider the propath separator is not the same as OS dependent file
**                           separator. The PROPATH_SEPARATOR string constant has been added to
**                           centralize this value. The fixed part of the PROPATH value has been
**                           initialized only once from server ini file and used in every propath
**                           modification as additional appendix after comma separator. The new
**                           change is adding the one more member to the WorkArea to separate the
**                           fixed and changeable parts. The member propath has been removed from
**                           main class, this value is getting now dynamically.
** 022 CA  20130604          Added setter for the batch mode state.
** 023 EVL 20130625          Changing the handling of incoming load() call to work with
**                           environment. In case of only two character or String parameters in
**                           this call we have environment and base-key input options. Because
**                           the directory option is useless in this case - this option is valid
**                           only for INI file based environment but without base-key "INI" the
**                           load() call is referred to the registry related case. For registry
**                           on the other hand the directory option can not be applied. So two
**                           input parameters means only environment and base-key variant.
** 024 CA  20140218          PROGRAM-NAME is now implemented by ProcedureManager.getStackEntry,
**                           which is called by getSourceName().
** 025 EVL 20140325          Changing request the OS name, type and path separator to the client
**                           side detection.
** 026 CA  20140702          Added isCustomEnvironment and getEnvironmentName APIs.
**                           Added generic's.  Removed unused WorkArea.pkgroot field.
** 027 CA  20140715          Fixed getTempDirectory - it needs to interrogate the P2J client.
** 028 MAG 20140822          Added getLegacyPlatform APIs.
** 029 MAG 20140908          Added COLOR-TABLE and FONT-TABLE notification.
** 030 MAG 20140912          INI files are available on all platforms regarding client type.
** 031 GES 20141030          Added CPINTERNAL.
** 032 SVL 20150114          For admin client use server-side OS version.
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.BaseSysTable;
import com.goldencode.p2j.ui.LogicalTerminal;
import com.goldencode.p2j.net.*;

/**
 * Progress compatible environment-related static utility methods.
 * <p>
 * This supports some common global variables that are called built-in
 * functions in Progress but which take no parameters, not even empty
 * parenthesis. These variables are read-only (they cannot be assigned),
 * except for <code>PROPATH</code> which can be modified as it can be in
 * Progress as well.
 * <p>
 * The following is the list of supported variables:
 * <p>
 * <pre>
 * Progress Variable                    Java Equivalent
 * ---------------------                ----------------------
 * current-language                     {@link #getCurrentLanguage}
 * dataservers                          {@link #getDataServerList}
 * gateways                             {@link #getDataServerList}
 * opsys                                {@link #getOperatingSystem}
 * progress                             {@link #getRuntimeType}
 * promsgs                              {@link #getMessageSource}
 * propath                              {@link #getSearchPath}
 * proversion                           {@link #getVersion}
 * </pre>
 * <p>
 * <pre>
 * Progress Function                    Java Equivalent
 * ---------------------                ----------------------
 * program-name                         {@link #getSourceName}
 * </pre>
 * <p>
* <pre>
 * System Handle:Attribute              Java Equivalent
 * -------------------------            ----------------------
 * session:batch-mode                   {@link #isBatchMode}       (get)
 * session:charset                      {@link #getCharset}        (get)
 * session:temp-directory               {@link #getTempDirectory}  (get)
 * </pre>
 */
public class EnvironmentOps
{
   /** Progress OPSYS string for Unix systems. */
   public static final String OS_UNIX = "UNIX";
   
   /** Progress OPSYS string for Windows systems. */
   public static final String OS_WIN = "WIN32";
   
   /** Progress PROPATH separator string for all systems. */
   public static final String PROPATH_SEPARATOR = ",";
   
   /** 
    * Maps known Java os.name system property entries to a given Progress 
    * OPSYS string.  Entries should be inserted in lowercase and all keys
    * should be lowercased before reading from the map.
    */
   public static final Map<String, String> osMap = new HashMap<>();
   
   /** The fixed default part of the PROPATH value. */
   private static String propathFixed = null;
   
   static
   {
      // unix variants
      osMap.put("linux",    OS_UNIX);
      osMap.put("aix",      OS_UNIX);
      osMap.put("solaris",  OS_UNIX);
      osMap.put("sunos",    OS_UNIX);
      osMap.put("hp-ux",    OS_UNIX);
      osMap.put("freebsd",  OS_UNIX);
      osMap.put("mac os x", OS_UNIX);
      
      // windows variants
      osMap.put("windows xp",   OS_WIN);
      osMap.put("windows 2000", OS_WIN);
      osMap.put("windows nt",   OS_WIN);
      osMap.put("windows me",   OS_WIN);
      osMap.put("windows 98",   OS_WIN);
      osMap.put("windows 95",   OS_WIN);
      
      // Initialize default PROPATH value
      propathFixed = fixupPropath(Utils.getDirectoryNodeString(null, "searchpath-fixed", ""));
   }
   
   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer(); 
   
   /**
    * Gets the name of the current language being used for the Progress
    * compatible environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/currentLanguage
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/currentLanguage
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/currentLanguage
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/currentLanguage
    * <p>
    * If no value is found via this lookup, then the default value of
    * "?" will be returned.
    *
    * @return   The name of the current language.
    */
   public static character getCurrentLanguage()
   {
      String lang = Utils.getDirectoryNodeString(null,
                                                 "currentLanguage",
                                                 "?");
      return new character(lang);
   }
   
   /**
    * Gets the list of the data servers available in Progress compatible
    * environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/dataServers
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/dataServers
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/dataServers
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/dataServers
    * <p>
    * If no value is found via this lookup, then the default value of
    * "PROGRESS,ORACLE,AS400,ODBC,MSS" will be returned.
    *
    * @return   The list of data servers.
    */
   public static character getDataServerList()
   {
      String def     = "PROGRESS,ORACLE,AS400,ODBC,MSS";
      String dserver = Utils.getDirectoryNodeString(null,
                                                    "dataServers",
                                                    def);
      return new character(dserver);
   }

   /**
    * Gets the name of the operating system being upon which the Progress
    * compatible environment is running.  The first attempt to resolve this
    * value will occur via a lookup in the directory.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/opsys/override
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/opsys/override
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/opsys/override
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/opsys/override
    * <p>
    * If no value is found via this lookup, then the Java system property
    * "os.name" is queried.  This is then translated via an optional
    * mapping table in the directory (that same paths and precedence as
    * above but ending in "opsys/mapping/&lt;lowercased_system_propert_name&gt;.
    * <p>
    * If no mapping is found, then an internal mapping table in this class
    * is consulted.  This table maps the following:
    * <p>
    * <pre>
    * os.name          Progress OPSYS
    * ---------------- ---------------
    * "linux"          "UNIX"
    * "aix"            "UNIX"
    * "solaris"        "UNIX"
    * "sunos"          "UNIX"
    * "hp-ux"          "UNIX"
    * "freebsd"        "UNIX"
    * "mac os x"       "UNIX"
    * "windows xp"     "WIN32"
    * "windows 2000"   "WIN32"
    * "windows nt"     "WIN32"
    * "windows me"     "WIN32"
    * "windows 98"     "WIN32"
    * "windows 95"     "WIN32"
    * </pre>
    * <p>
    * If there is still no match, then any <code>os.name</code> with the
    * lowercase text "unix" will return "UNIX" and any <code>os.name</code>
    * with the lowercase text "windows" will return "WIN32". 
    * <p>
    * If there is still no match, the <code>os.name</code> returned by the
    * Java will be returned. 
    *
    * @return   The name of the current operating system.
    */
   public static character getOperatingSystem()
   {
      WorkArea wa = work.obtain();
      
      if (wa.clientOsName == null)
      {
         // look for a hard coded value
         String osname = Utils.getDirectoryNodeString(null,
                                                      "opsys/override",
                                                      "");
                                                      
         // if there is a hard coded operating system override, use it
         if (osname.length() == 0)
         {
            // query the JVM value (it will always be present)
            String fallback = null;

            // We have client running.
            // If the client is in the headless mode (e.g. admin client), there is no physical
            // terminal, so the OS can't be accessed.
            if (!TransactionManager._isHeadless() && wa.osInsp != null)
            {
               fallback = wa.osInsp.getOperatingSystem();
            }
            // We make conversion, take standalone/server value
            else
            {
               fallback = System.getProperty("os.name");
            }
            // New candidate               
            String sysprop  = fallback.toLowerCase();
         
            // check if there is a mapping in the directory for this exact
            // value
            osname = "opsys/mapping/" + sysprop;
            osname = Utils.getDirectoryNodeString(null, osname, "");
                                                   
            // use the mapping if there was one
            if (osname.length() == 0)
            {
               // attempt to translate it by using our predefined tables
               osname = (String) osMap.get(sysprop);
            
               // no match?
               if (osname == null)
               {
                  // last try
                  if (sysprop.indexOf("unix") != -1)
                  {
                     osname = OS_UNIX;
                  }
                  else if (sysprop.indexOf("windows") != -1)
                  {
                     osname = OS_WIN;
                  }
                  else
                  {
                     // all else fails, return what Java specified
                     osname = fallback;
                  }
               }
            }
         }

         // Store per session OS name string to avoid every time call down to client
         wa.clientOsName = osname;

         // Calculate the cache for boolean Windows flag
         wa.isWindowsClient = osname.toLowerCase().indexOf("win") != -1; 
      }
      
      return new character(wa.clientOsName);
   }
   
   /**
    * Checks is the current OS the P2J is running on belongs to the Windows
    * family.
    *
    * @return   <code>true</code> if the current OS is from Windows family or 
    *           <code>false</code> otherwise.
    */
   public static boolean isUnderWindowsFamily()
   {
      WorkArea wa = work.obtain();
      
      if (wa.clientOsName == null)
      {
         getOperatingSystem();
      }
      
      return wa.isWindowsClient;
   }
   
   /**
    * Gets the current runtime type being provided by the Progress
    * compatible environment.  The P2J environment only provides a
    * "Full" environment.
    *
    * @return   Always returns the string "Full".
    */
   public static character getRuntimeType()
   {
      return new character("Full");
   }
   
   /**
    * Gets the name of the directory where temporary files are stored.
    * Corresponds with with the <code>SESSION:TEMP-DIRECTORY</code> system
    * handle attribute.
    * 
    * @return  Name of temporary directory.
    */
   public static character getTempDirectory()
   {
      return new character(work.get().osInsp.getTempDirectory());
   }
   
   /**
    * Gets the name of the current message source being used for the Progress
    * compatible environment.
    *
    * @return   Always returns the string "promsgs".
    */
   public static character getMessageSource()
   {
      return new character("promsgs");
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    * @param    conversion
    *           This marks if the function is called by Preprocessor
    *           at conversion time or not
    */
   public static void setSearchPath(String path , boolean conversion)
   {
      if (path != null)
      {
         // The new variable PROPATH part
         String currentOverride = fixupPropath(path);

         // Combine the final value from variable and fixed parts         
         String currentPropath = currentOverride;
         if (!propathFixed.equals(""))
         {
            currentPropath = currentPropath + PROPATH_SEPARATOR + propathFixed;
         }

         // Replace old local PROPATH values with new one, the full effective value 
         work.obtain().propath = currentPropath;
         // Only changeable part of the PROPATH
         work.obtain().propathOverride = currentOverride;

         if (!conversion)
         {
            // Use local resuting propath to set run time value
            String[] pathList = splitPath(currentPropath, PROPATH_SEPARATOR);
            FileSystemOps.setSearchPath(pathList, getLegacyCaseSensitive());
         }
      }
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    */
   public static void setSearchPath(String path)
   {
      setSearchPath(path, false);
   }
   
   /**
    * Sets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * assignment of the "PROPATH" variable in Progress and represents a
    * mechanism to override the default that is available from the directory.
    * <p>
    * The PROPATH is also used in the implementation of
    * {@link FileSystemOps#searchPath}.  The actual I/O operations when
    * executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    *
    * @param    path
    *           The replacement path to Progress source files.
    */
   public static void setSearchPath(character path)
   {
      setSearchPath(path.toStringMessage());
   }
   
   /**
    * Gets any context-specific override for the Progress search path which
    * was previously set via {@link #setSearchPath}.
    *
    * @return   The path override or <code>null</code> if no override was
    *           ever set.
    */
   public static String getSearchPathOverride()
   {
      return work.obtain().propath;
   }
   
   /**
    * Gets the Progress search path on the TARGET P2J system which is
    * used for the Progress compatible environment.  This corresponds to the
    * use of the "PROPATH" variable in Progress and is also used in the
    * implementation of {@link FileSystemOps#searchPath}.  The actual I/O
    * operations when executed via the Progress compatible interface (see
    * {@link FileSystemOps} and {@link Stream}) for a P2J system are executed
    * on the P2J client.  This is the TARGET system on which this path is
    * valid.
    * <p>
    * This must be distinguished from the original Progress SOURCE system's
    * PROPATH which must be maintained for certain runtime translations
    * of dynamically generated names (the algorithms of which were designed
    * with that path assumed) but which is not necessarily a valid path on 
    * the system on which this is running.  The original PROPATH can be
    * obtained via the {@link #getLegacySearchPath} method.
    * <p>
    * If a specific PROPATH has been assigned in this user's context, then
    * that "override" will be returned.  Otherwise, the directory will be
    * consulted for the stored version.
    * <p>
    * The value returned from the directory may have been found via a search
    * algorithm that is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/searchpath
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/searchpath
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/searchpath
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/searchpath
    * <p>
    * If no value is found via this lookup, then the default value of
    * ".," will be returned.
    *
    * @return   The path to Progress source files.
    */
   public static character getSearchPath()
   {
      WorkArea wa      = work.obtain();
      String   propath = wa.propath;
      
      // is there a context-specific override?
      if (propath == null)
      {
         // support the override
         propath = wa.propathOverride;
         // There is system default, add to the end
         if (!propathFixed.equals(""))
         {
            propath = propath + PROPATH_SEPARATOR + propathFixed;  
         }
         // Save the local copy of the effective propath
         wa.propath = propath; 
      }
      
      return new character(propath);
   }
   
   /**
    * Substitutes the path separator values with the comma character in
    * the given string and returns the 4GL compatible PROPATH value that is
    * the combination of the directory entries separated by comma.
    *
    * @param    propathInitial
    *           The PROPATH that possibly needs to be fixed due to containing
    *           of the path separators instead of the expected comma values.
    *
    * @return   The PROPATH value with comma separation instead of possible
    *           path separators.
    */
   public static String fixupPropath(String propathInitial)
   {
      // First select the right path separator
      String pathSep = Utils.getDirectoryNodeString(null, "path-separator", File.pathSeparator);
      StringBuffer sb = new StringBuffer();
      int pathSepLength = pathSep.length();
      
      // The starting position of the next PROPATH element
      int nextSegment = 0;
      // Scan input parameter for the next occurence of the path separator.
      // The next PROPATH element starting point is the sum of the current
      // PROPATH element end point and length of the path sepatator.
      for (int i = 0;
           (i = propathInitial.indexOf(pathSep, nextSegment)) >= 0;
           nextSegment = i + pathSepLength)
      {
         // Extract the PROPATH entry up to the nearest path separator 
         sb.append(propathInitial.substring(nextSegment, i));
         // Replace path separator with comma
         sb.append(PROPATH_SEPARATOR);
      }
      // We have to handle the case when new PROPATH sting is not finished
      // with path separator.
      if (nextSegment < propathInitial.length())
         sb.append(propathInitial.substring(nextSegment));
      
      return sb.toString();
   }
   
   /**
    * Gets the original Progress SOURCE system's PROPATH which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/propath
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/propath
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/propath
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/propath
    * <p>
    * If no value is found via this lookup, then the default value of
    * "." will be returned.
    * <p>
    * This must be distinguished from Progress source file/search path on the
    * TARGET P2J system which is used for the Progress compatible
    * environment.  To obtain that value, use {@link #getSearchPath}.
    *
    * @return   The path to Progress source files.
    */
   public static String getLegacySearchPath()
   {
      return Utils.getDirectoryNodeString(null, "file-system/propath", ".");
   }
   
   /**
    * Gets the path separator character used to parse the Progress search
    * path (see {@link #getSearchPath}).
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/path-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/path-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/path-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/path-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * ':' will be returned.
    *
    * @return   The path to Progress source files.
    */
   public static String getPathSeparator()
   {
      if (work == null)
      {
         return Utils.getDirectoryNodeString(null, "path-separator", File.pathSeparator);
      }
      
      WorkArea wa = work.obtain();
      
      if (wa.clientPathSeparator == null)
      {
         // First check the directory node first
         String pathSep = Utils.getDirectoryNodeString(null, "path-separator", "");
         
         if (pathSep.length() == 0 && wa.osInsp != null)
         {
            // Request client side as fallback for path separator value
            pathSep = wa.osInsp.getPathSeparator();
         }
         // Last try
         if (pathSep == null || pathSep.length() == 0)
         {
            pathSep = File.pathSeparator;
         }
         // But check the directory node first
         wa.clientPathSeparator = pathSep;
      }
      
      return wa.clientPathSeparator;
   }
   
   /**
    * Gets the original Progress SOURCE system's path separator which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/path-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/path-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/path-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/path-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * ":" will be returned.
    * <p>
    * This must be distinguished from the path separator on the (runtime)
    * TARGET P2J system which is used for the Progress compatible
    * environment.  To obtain that value, use {@link #getPathSeparator}.
    *
    * @return   The path to Progress source files.
    */
   public static String getLegacyPathSeparator()
   {
      return Utils.getDirectoryNodeString(null,
                                          "file-system/path-separator",
                                          File.pathSeparator);
   }
   
   /**
    * Gets the list of directories in the Progress source file path.  Each
    * element of the array is a directory in the source path.  This source
    * path is split based on the path separator used in this path
    * specification which may be different than that reported by the
    * system (<code>File.pathSeparator</code>.  The resulting array is
    * returned to the caller.
    * <p>
    * The {@link #getSearchPath} is used for the path specification and
    * the {@link #getPathSeparator} is used to split the path.
    *
    * @return   The array of directory names.
    */
   public static String[] getSearchPathList()
   {
      return splitPath(getSearchPath().toStringMessage(), PROPATH_SEPARATOR);
   }
   
   /**
    * Gets the original Progress SOURCE system's file separator which must be
    * maintained for certain runtime translations of dynamically generated
    * names (the algorithms of which were designed with that path assumed)
    * but which is not necessarily a valid path on the system on which this
    * is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/file-separator
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/file-separator
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/file-separator
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/file-separator
    * <p>
    * If no value is found via this lookup, then the default value of
    * "/" will be returned.
    *
    * @return   The legacy system's file separator.
    */
   public static String getLegacyFileSeparator()
   {
      return Utils.getDirectoryNodeString(null,
                                          "file-system/file-separator",
                                          File.separator);
   }
   
   /**
    * Gets the original Progress SOURCE file system's case-sensitive setting
    * which must be maintained for certain runtime translations of dynamically
    * generated names (the algorithms of which were designed with that path
    * assumed) but which is not necessarily a valid path on the system on
    * which this is running.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/file-system/case-sensitive
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/file-system/case-sensitive
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/file-system/case-sensitive
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/file-system/case-sensitive
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>true</code> will be returned.
    *
    * @return   <code>true</code> if the legacy file system was 
    *           case-sensitive.
    */
   public static boolean getLegacyCaseSensitive()
   {
      return Utils.getDirectoryNodeBoolean(null,
                                           "file-system/case-sensitive",
                                           !isUnderWindowsFamily());
   }
   
   /**
    * Gets the current version string of the Progress compatible environment.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/version
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/version
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/version
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/version
    * <p>
    * If no value is found via this lookup, then the default value of
    * "10.2B" will be returned.
    *
    * @return   The current version string.
    */
   public static character getVersion()
   {
      String version = Utils.getDirectoryNodeString(null,
                                                    "version",
                                                    "10.2B");
      return new character(version);
   }
   
   /**
    * Returns the Progress source file name that matches the program at the given index in the 
    * stack trace if the given index position exists. The Progress source file name will be 
    * returned if possible using the {@link ProcedureManager#getStackEntry}.
    *
    * @param    index
    *           The 1-based index position in the call stack to be returned.
    *
    * @return   The Progress source file name (if the class has the proper member).  If the index
    *           is out of bounds, then the <code>unknown value</code> is returned.
    */
   public static character getSourceName(integer index)
   {
      if (index.isUnknown())
      {
         return new character();
      }
      
      return getSourceName(index.intValue());
   }
   
   /**
    * Returns the Progress source file name that matches the program at the given index in the 
    * stack trace if the given index position exists. The Progress source file name will be 
    * returned if possible using the {@link ProcedureManager#getStackEntry}.
    *
    * @param    index
    *           The 1-based index position in the call stack to be returned.
    *
    * @return   The Progress source file name (if the class has the proper member).  If the index
    *           is out of bounds, then the <code>unknown value</code> is returned.
    */
   public static character getSourceName(int index)
   {
      return new character(ProcedureManager.getStackEntry(index - 1));
   }
   
   /**
    * Splits a path using the alternative separator string
    * 
    * @param path 
    *        The <code>String</code> path to split into an array of directory
    *        names.
    * @param separator 
    *        The <code>String</code> separator to split into an array of directory
    *        names.
    * 
    * @return The array of directory names.
    */
   private static String[] splitPath(String path, String separator)
   {
      if (separator == null)
      {
         separator = getPathSeparator();
      }
      if (path != null)
      {
         return path.split(separator, -1);
      }
      else
      {
         return null;
      }
   }
   
   /**
    * Reports if the current context is running in batch mode.
    * <p>
    * The value returned may have been found via a search algorithm that
    * is account (user or process) specific or group specific
    * within the current server or a global default for all servers. 
    * <p>
    * The implementation iteratively looks up the directory node under:
    * /server/&lt;serverID&gt;/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * If no user/process or group nodes are present, then this is checked:
    * /server/&lt;serverID&gt;/runtime/default/batchMode
    * <p>
    * If no /server/&lt;serverID&gt;/runtime node exists, this is checked
    * (it is the global default area for all servers):
    * /server/default/runtime/&lt;account_or_group&gt;/batchMode
    * <p>
    * Finally, if no user/process or group nodes are present in the global
    * default area, then this is checked:
    * /server/default/runtime/default/batchMode
    * <p>
    * If no value is found via this lookup, then the default value of
    * <code>false</code> will be returned.
    *
    * @return   <code>true</code> if the current context is in batch mode.
    */
   public static logical isBatchMode()
   {
      WorkArea wa = work.obtain();
      if (wa.batchMode == null)
      {
         wa.batchMode = Utils.getDirectoryNodeBoolean(null, "batchMode", false);
      }
      
      return new logical(wa.batchMode);
   }
   
   /**
    * Force the batch mode of the current context to the specified state.
    *
    * @param   batchMode
    *          <code>true</code> if this context must be in batch mode.
    */
   public static void setBatchMode(boolean batchMode)
   {
      WorkArea wa = work.obtain();
      wa.batchMode = batchMode;
   }
   
   /**
    * Reports the default charset used by the JVM. Since the Progress default charset is
    * ISO-8859-1, when the JVM reports that UTF-8 is the default, this method will return
    * the string "ISO-8859-1" instead.
    * <p>
    * If the 4GL environment is actually configured for UTF-8, then this approach won't work.
    * If the default codepage name for the JVM is not UTF-8, it is assumed that the codepage
    * has been overriden already (at JVM startup).  In that case, the value reported by the JVM
    * is returned unchanged.
    *
    * @return   The default charset name.
    */
   public static character getCharset()
   {
      // TODO: is this right?
      return getCPInternal();
   }

   /**
    * Reports the default codepage used by the JVM. Since the Progress default charset is
    * ISO8859-1, when the JVM reports that UTF-8 is the default, this method will return
    * the string "ISO8859-1" instead.
    * <p>
    * If the 4GL environment is actually configured for UTF-8, then this approach won't work.
    * If the default codepage name for the JVM is not UTF-8, it is assumed that the codepage
    * has been overriden already (at JVM startup).  In that case, the value reported by the JVM
    * is returned unchanged.
    *
    * @return   The codepage used for internal operations.
    */
   public static character getCPInternal()
   {
      String cp = Utils.getCharsetName();
      
      if ("ISO-8859-1".equals(cp))
      {
         // TODO: the 4GL doesn't insert the extra hyphen, this really needs to be explored
         cp = "ISO8859-1";
      }
      
      return new character(cp);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    */
   public static void load(character env)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(false), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    */
   public static void load(String env)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(false), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(character env, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(String env, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, boolean flagNew, String baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, boolean flagNew, String baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(false), baseKey);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(false), baseKey);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, String baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, (character)null, new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, String baseKey)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), (character)null, new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(character env, character directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, directory, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(String env, character directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), directory, new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(character env, String directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(env, new character(directory), new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    */
   public static void load(String env, String directory, boolean flagNew)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      load(new character(env), new character(directory), new logical(flagNew), (character)null);
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, String baseKey)
   {
      load(env, directory, new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, String directory, String baseKey)
   {
      load(env, new character(directory), new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character directory, String baseKey)
   {
      load(new character(env), directory, new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, String directory, String baseKey)
   {
      load(new character(env), new character(directory),
           new logical(false), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, boolean flagNew, String baseKey)
   {
      load(env, directory, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, String directory, boolean flagNew, String baseKey)
   {
      load(env, new character(directory), new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, character directory, boolean flagNew, String baseKey)
   {
      load(new character(env), directory, new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(String env, String directory, boolean flagNew, String baseKey)
   {
      load(new character(env), new character(directory),
           new logical(flagNew), new character(baseKey));
   }

   /**
    * Creates application defaults implementing Progress LOAD statement.
    * 
    * @param env 
    *        The environment to create or load. Can be the registry key,
    *        initialization file.
    * @param directory 
    *        The directory path for existing initialization file or place
    *        to create initialization file.
    * @param flagNew
    *        If <code>true</code> - the new file or registry key will be
    *        created overwriting the possibly existed one.
    * @param baseKey
    *        When the parameter is specified the serching is performed under
    *        the particular key value. It is possible to specify "INI" as
    *        parameter. In this case search or creation will be dome inside
    *        initialization file specified in env parameter.
    */
   public static void load(character env, character directory, logical flagNew, character baseKey)
   {
      // check if INI file
      boolean isIni = EnvironmentReader.TYPE_INI.equalsIgnoreCase(baseKey.toStringMessage());
            
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIni)
      {
         onLoadError(env, directory);
         return;
      }
      
      // Preparing initial parameters for EnvironmetsAccessor interface
      String envAsString = null;
      String directoryAsString = null;
      String baseKeyAsString = null;
      boolean flagNewAsBoolean = false;
      
      if (env != null)
      {
         envAsString = env.toStringMessage();
      }
      if (directory != null)
      {
         directoryAsString = directory.toStringMessage();
      }
      if (flagNew != null)
      {
         flagNewAsBoolean = flagNew.booleanValue();
      }
      if (baseKey != null)
      {
         baseKeyAsString = baseKey.toStringMessage();
      }

      if (!Environments.load(envAsString, directoryAsString, flagNewAsBoolean, baseKeyAsString))
      {   
         onLoadError(env, directory);
      }
   }

   /**
    * On load error message.
    * 
    * @param    env 
    *           The environment to create or load. Can be the registry key,
    *           initialization file.
    * @param    directory 
    *           The directory path for existing initialization file or place
    *           to create initialization file.
    */           
   private static void onLoadError(character env, character directory)
   {
      // This is support for NO-ERROR option
      if (directory == null || directory.isUnknown())
      {
         ErrorManager.conditionalShowErrorAndAbend(4450, "The LOAD of "+env.toStringMessage()+
                                                         " failed");
      }
      else
      {
         ErrorManager.conditionalShowErrorAndAbend(4449, "The LOAD of "+env.toStringMessage()+
                                                         " in path "+
                                                         directory.toStringMessage()+
                                                         " failed");
      }      
   }
   
   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is empty string ("") the default
    *        environment becomes the current one.
    */
   public static void use(String env)
   {
      use(new character(env));
   }

   /**
    * Specifies environment defaults to use in subsequent windows.
    * 
    * @param env 
    *        The environment to load. If the parameter is empty string ("") the default
    *        environment becomes the current one.
    */
   public static void use(character env)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         ErrorManager.conditionalShowErrorAndAbend(4452, "The USE of "+env.toStringMessage()+
                                                         " failed");
         return;
      }
      
      String envAsString = null;

      if (env != null)
      {
         envAsString = env.toStringMessage();
      }
      
      Environments.use(envAsString);

      // notify
      envAsString = getCurrentEnvironmentName();
      onUse(environmentName(envAsString));
   }
   
   /**
    * Check if a custom environment is in use.
    * 
    * @return   See above.
    */
   public static boolean isCustomEnvironment()
   {
      return (isUnderWindowsFamily() || isIniType()) 
               && Environments.isCustomEnvironment();
   }
   
   /**
    * Get the current environment's name.
    * 
    * @return   See above.
    */
   public static String getEnvironmentName()
   {
      return Environments.getEnvironmentName();
   }

   /**
    * Get the current environment type.
    * 
    * @return   See above.
    */
   public static String getEnvironmentType()
   {
      return Environments.getEnvironmentType();
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, String value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(section, (character)null, new character(value));
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, character value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(new character(section), (character)null, value);
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(new character(section), (character)null, new character(value));
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, character value)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      setKeyValue(section, (character)null, value);
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, character key, String value)
   {
      setKeyValue(section, key, new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, String key, character value)
   {
      setKeyValue(section, new character(key), value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, character key, character value)
   {
      setKeyValue(new character(section), key, value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, String key, String value)
   {
      setKeyValue(section, new character(key), new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String key, character value)
   {
      setKeyValue(new character(section), new character(key), value);
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, character key, String value)
   {
      setKeyValue(new character(section), key, new character(value));
   }
   
   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(String section, String key, String value)
   {
      setKeyValue(new character(section), new character(key), new character(value));
   }

   /**
    * Adds, modifies and deletes keys in current environment.
    * 
    * @param section 
    *        The name of the section containing the key to modify.
    * @param key 
    *        The name of the key key to modify or default key if not specified.
    * @param value 
    *        The new value of the key under modification.
    */
   public static void setKeyValue(character section, character key, character value)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         return;
      }
      
      // Preparing initial parameters for EnvironmetsAccessor interface
      String sectionAsString = null;
      String keyAsString = null;
      String valueAsString = null;

      if (section != null)
      {
         // Unknown section value should be handled the same way as empty string
         if (section.isUnknown())
         {
            sectionAsString = "";
         }
         else
         {
            sectionAsString = section.toStringMessage();
         }
      }
      if (key != null)
      {
         // Unknown key value should be handled the same way as empty string
         if (key.isUnknown())
         {
            keyAsString = "";
         }
         else
         {
            keyAsString = key.toStringMessage();
         }
      }
      if (value != null)
      {
         // Unknown value should be handled the same way as empty string
         if (value.isUnknown())
         {
            valueAsString = "";
         }
         else
         {
            valueAsString = value.toStringMessage();
            if (valueAsString.equals("?"))
            {
               valueAsString = "";
            }
         }
      }

      Environments.setKeyValue(sectionAsString, keyAsString, valueAsString);
   }

   /**
    * Getting the default key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(character section)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      return getKeyValue(section, (character)null);
   }

   /**
    * Getting the default key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(String section)
   {
      // Using null as missing parameter means the usage the default
      // value for parameter, not the same as unknown. Because the
      // Progress unknown value is the legal value and can be passed
      // withing not-null variable instance.
      return getKeyValue(new character(section), (character)null);
   }

   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(String section, String key)
   {
      return getKeyValue(new character(section), new character(key));
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(character section, String key)
   {
      return getKeyValue(section, new character(key));
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(String section, character key)
   {
      return getKeyValue(new character(section), key);
   }
   
   /**
    * Getting the key value from the current environment.
    * 
    * @param section 
    *        The name of the section containing the key to get.
    * @param key 
    *        The name of the key key to get or default key if not specified.
    * 
    * @return The value of the key.
    */
   public static character getKeyValue(character section, character key)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         return new character();
      }
      
      // Preparing initial parameters for EnvironmetsAccessor interface
      String sectionAsString = null;
      String keyAsString = null;

      if (section != null)
      {
         // Unknown section value should be handled the same way as empty string
         if (section.isUnknown())
         {
            sectionAsString = "";
         }
         else
         {
            sectionAsString = section.toStringMessage();
         }
      }
      if (key != null)
      {
         // Unknown key value should be handled the same way as empty string
         if (key.isUnknown())
         {
            keyAsString = "";
         }
         else
         {
            keyAsString = key.toStringMessage();
         }
      }
      
      return new character(Environments.getKeyValue(sectionAsString, keyAsString));
   }

   /**
    * Unloads specifies environment from the current one.
    * 
    * @param env 
    *        The environment to unload. Can be the registry or initialization
    *        file. The environment to unload should be previously loaded by
    *        load() method.
    */
   public static void unload(String env)
   {
      unload(new character(env));
   }
   
   /**
    * Unloads specified environment from the current one if possible.
    * 
    * @param env 
    *        The environment to unload. Can be the registry or initialization
    *        file. The environment to unload should be previously loaded by
    *        load() method.
    */
   public static void unload(character env)
   {
      // This subsystem works only under Windows family OS
      // INI files works on all platforms
      if (!isUnderWindowsFamily() && !isIniType())
      {
         String msg = "The UNLOAD of " + env.toStringMessage() + " failed";
         ErrorManager.conditionalShowErrorAndAbend(4451, msg);
         return;
      }
      
      String envAsString = null;

      if (env != null)
      {
         envAsString = env.toStringMessage();
      }

      // check if unload is possible
      if (LogicalTerminal.canUnloadEnvironment(environmentName(envAsString)))
      {
         Environments.unload(envAsString);
         onUnload(environmentName(envAsString));
      }
      else
      {
         String msg = "The UNLOAD of " + envAsString + " failed";
         ErrorManager.conditionalShowErrorAndAbend(4451, msg);
      }
   }
   
   /**
    * Get current environment name. If no custom environment is set return default environment
    * name which is a <code>null</code> value.
    * 
    * @return  see above.
    */
   public static String getCurrentEnvironmentName()
   {
      String envName = null;
      
      if (isCustomEnvironment())
      {
         envName = getEnvironmentName();
      }
      
      return environmentName(envName);
   }
   
   /**
    * Get client legacy platform name.
    * <p> 
    * Sometimes we need to specify a legacy platform regarding the real platform on which the 
    * clients actually runs. @link ColorSpec#convert. By default the client legacy platform is 
    * the same as OPSYS built-in returned value in Progress.
    * <p> 
    * The value is read from server directory. If no value is set for <code>legacyPlatform</code>
    * then the returned value is the same as returned by the <code>getOperatingSystem</code> call.
    * If set in the directory, the value of legacyPlatform MUST be UNIX or WIN32.
    * 
    * @return  Client legacy platform name:
    *          UNIX  - Unix/Linux OS familly.
    *          WIN32 - Windows OS familly.          
    */
   public static String getLegacyPlatform()
   {
      WorkArea wa = work.obtain();
      
      if (wa.legacyPlatform == null)
      {
         String os = getOperatingSystem().toStringMessage();
         wa.legacyPlatform = Utils.getDirectoryNodeString(null, "legacyPlatform", os);
      }
      
      return wa.legacyPlatform;
   }

   /**
    * Resolve environment name.  Convert to lower case to be compliant with FONT-TABLE
    * 
    * @param   envName
    *          Environment name.
    *          
    * @return  see above.
    */
   private static String environmentName(String envName)
   {
      return (envName == null || envName.trim().isEmpty()) ? null : envName.toLowerCase();      
   }

   /**
    * A notification when <code>USE environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   private static void onUse(String envName)
   {
      BaseSysTable.use(envName);
   }

   /**
    * A notification when <code>UNLOAD environment</code> statement is executed.
    * 
    * @param   envName
    *          Environment name.
    */
   private static void onUnload(String envName)
   {
      BaseSysTable.unload(envName);
   }

   /**
    * Check if it's a INI file environment type. 
    *  
    * @return  <code>true</code> if INI file environment type.
    */
   private static boolean isIniType()
   {
      return EnvironmentReader.TYPE_INI.equals(getEnvironmentType());
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Stores any context-specific full effective override to the <code>PROPATH</code>. */
      private String propath = null;
      
      /** Stores any context-specific override to the <code>PROPATH</code>, changeable part. */
      private String propathOverride =
                    fixupPropath(Utils.getDirectoryNodeString(null, "searchpath", "."));
      
      /** The batch-mode state of this context. */
      private Boolean batchMode = null;
      
      /** The name of the OS on the client side. */
      private String clientOsName = null;
   
      /** The flag indicating Windows on client side. */
      private boolean isWindowsClient;
   
      /** The string representing delimiter between path entries on the client side. */
      private String clientPathSeparator = null;
   
      /** Object to check OS properties on client side. */
      private OperatingSystemInspector osInsp = null;
      
      /** Client legacy plaform name */
      private String legacyPlatform = null;      
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();

         // This will work for run-time only
         if (SessionManager.get() != null)
         {
            wa.osInsp = (OperatingSystemInspector) RemoteObject.obtainInstance(
                                                       OperatingSystemInspector.class, true);
         }
         
         return wa;
      }   
   }   
}
